#include <pspdebug.h>
#include <stdlib.h>
#include <string.h>
#include "buffer.h"

#define LINEINCREMENT 16

struct TextLine *newTextLine(){
	/* Well, no mistery about this function, only initializes
	 * a new line */
	struct TextLine *line;
	if (!(line = malloc(sizeof(struct TextLine)))) exit(-1);
	if (!(line->text = malloc(LINEINCREMENT*sizeof(char)))) exit(-1);
	line->length = 0; // Keeps track of were is the end of the line, pointing always to the '\0' character
	line->allocable = LINEINCREMENT - 1;
	line->text[line->length] = '\0';
	line->next = NULL;
	line->last = NULL;
	return line;
}

int lineExtend(struct TextLine *line){
	/* Reallocates the text from a line with space for LINEINCREMENT
	 * more characters */
	if (!(line->text = realloc(line->text,(line->allocable + LINEINCREMENT)*sizeof(char)))) exit(-1);
	line->allocable = line->allocable + LINEINCREMENT;
	return 1;
}

int lineAddChar(struct TextLine *line, char newchar, int position){
	/* Writes a character at a given position of a line, displacing forwards
	 * all characters after that position (should be rewritten better)*/
	char *tmpstr;
	struct TextLine *newline;
	
	// Next four lines store the part of the string from the writing position in tmpstr, then write the desired character (even '\n') on the designed position
	if (line->length >= line->allocable - 1) lineExtend(line);
	if (!(tmpstr = malloc((line->allocable - position + 1)*sizeof(char)))) exit(-1);
	strcpy(tmpstr,line->text + position); // Copy the part of the line after the cursor
	line->text[position] = newchar;
	// If the written character is not '\n' we copy tmpstr after the position and increment the line length
	if (newchar != '\n'){
		strcpy(line->text + position + 1, tmpstr);
		(line->length)++; // The '\n' character substitutes '\0' as the end of the string and the length continues the same, important for the cursor to work cleanly
	}
	else{ // Messy part for writing '\n' (and thus inserting lines)
		line->text[position + 1] = '\0'; // Write the ending '\0' of the divided line
		newline = newTextLine(); // 3 lines, write the string after the writing position into a new line
		while (newline->allocable < line->length - position + 1) lineExtend(line->last); // Not sure
		strcpy(newline->text, tmpstr);
		newline->length = line->length - position;
		line->length = position; // Remember that length is the position of the first non-writable character
		newline->next = line->next; // 4 lines, link the line list
		newline->last = line;
		if (line->next) line->next->last = newline;
		line->next = newline;
	}
	free(tmpstr); // Phew!, maybe it works flawlessly?
	return 1;
}

int lineDelChar(struct TextLine *line, int position){
	/* Deletes the character before given position, if position is 0 the
	 * entire line is appended at the end of the previous line */
	char *tmpstr;
	
	if (!(tmpstr = malloc((line->allocable - position + 1)*sizeof(char)))) exit(-1);
	strcpy(tmpstr,line->text + position);
	
	if (position == 0 && line->last){
		while (line->last->allocable < line->last->length + line->length + 2) lineExtend(line->last);
		strcpy(line->last->text + line->last->length,tmpstr);
		line->last->length = line->last->length + line->length;
		line->last->next = line->next;
		if (line->next) line->next->last = line->last;
		free(line->text);
		free(line);
	}
	else {
		strcpy(line->text + position - 1, tmpstr);
		(line->length)--;
	}
	return 1;
}

struct TextBuffer *newTextBuffer(){
	/* Initializes a new buffer, fairly simple, isn't it? */
	struct TextBuffer *buffer;
	if (!(buffer = malloc(sizeof(struct TextBuffer)))) exit(-1);
	buffer->lines = newTextLine();
	buffer->cursorline = buffer->lines;
	buffer->cursorcolumn = 0;
	return buffer;
}

void destroyBuffer(struct TextBuffer *buffer){
	/* Frees all memory used by a buffer and its lines */
	struct TextLine *line;
	while (buffer->lines){
		line = buffer->lines;
		buffer->lines = line->next;
		free(line);
	}
	free(buffer);
}

// There will be a BufferAddLine function that inserts a previously created line in a given position
int bufferNewLine(struct TextBuffer *buffer){
	/* Creates a new line and links it to the cursor line of the given
	 * buffer DEPRECATED, now bufferAddChar(buffer, '\n') should be used */
	struct TextLine *line;
	//bufferAddChar(buffer,'\n');
	line = newTextLine();
	line->next = buffer->cursorline->next;
	line->last = buffer->cursorline;
	buffer->cursorline->next = line;
	buffer->cursorline = line;
	buffer->cursorcolumn = 0;
	return 1;
}

int bufferAddChar(struct TextBuffer *buffer, char newchar){
	/* Adds a character to the end of a line and moves the cursor */
	lineAddChar(buffer->cursorline,newchar,buffer->cursorcolumn);
	if (newchar != '\n') (buffer->cursorcolumn)++;
	else {
		buffer->cursorline = buffer->cursorline->next;
		buffer->cursorcolumn = 0;
	}
	return 1;
}

int bufferDelChar(struct TextBuffer *buffer){
	int finalcolumn;
	
	if (buffer->cursorcolumn == 0){
		if (buffer->cursorline->last){ // Which means that we aren't in the first line
			buffer->cursorline = buffer->cursorline->last;
			finalcolumn = buffer->cursorline->length;
			lineDelChar(buffer->cursorline->next,buffer->cursorcolumn); // The line introduced will be freed, so it's important to be sure that buffer->cursorline don't point it.
			buffer->cursorcolumn = finalcolumn;
		}
	}
	else {
		lineDelChar(buffer->cursorline,buffer->cursorcolumn);
		buffer->cursorcolumn--;
	}
	return 1;
}

int moveCursorRight(struct TextBuffer *buffer){
	if (buffer->cursorcolumn < buffer->cursorline->length) (buffer->cursorcolumn)++;
	return buffer->cursorcolumn;
}

int moveCursorUp(struct TextBuffer *buffer){
	if (buffer->cursorline != buffer->lines) buffer->cursorline = buffer->cursorline->last;
	if (buffer->cursorcolumn > buffer->cursorline->length) buffer->cursorcolumn = buffer->cursorline->length;
	return buffer->cursorcolumn;
}

int moveCursorLeft(struct TextBuffer *buffer){
	if (buffer->cursorcolumn > 0) (buffer->cursorcolumn)--;
	return buffer->cursorcolumn;
}

int moveCursorDown(struct TextBuffer *buffer){
	if (buffer->cursorline->next) buffer->cursorline = buffer->cursorline->next;
	if (buffer->cursorcolumn > buffer->cursorline->length) buffer->cursorcolumn = buffer->cursorline->length;
	return buffer->cursorcolumn;
}
