#include <pspkernel.h>
#include <pspdebug.h>
#include <pspdisplay.h>
#include <pspctrl.h> 
#include <stdlib.h>

#include <math.h>

#include "input.h"
#include "buffer.h"
#include "display.h"
#include "command.h"

PSP_MODULE_INFO("PSPTed", 0, 1, 1);

/* Exit callback */
int exit_callback(int arg1, int arg2, void *common) {
	sceKernelExitGame();
	return 0;
}

/* Callback thread */
int callbackThread(SceSize args, void *argp) {
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int setupCallbacks(void) {
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", callbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0) {
	sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

int main(){
	/* This is the main function, right now, it englobes way too much code,
	 * also, is not good code. I test new functions here first and after
	 * that I move them to modules. */
	// Declare variables
	SceCtrlData pad;
	int c = 0, oldc = -1; // c is the character returned from pad, searching for a better name...
	int commandmode = 0;
	int charset = 0;
	int cursorx = 0, cursory = 0;
	struct TextBuffer *textbuffer,*activebuffer,*commandbuffer=NULL;
	
	// Set PSP parameters
	sceCtrlSetSamplingMode(1);
	setupCallbacks(); 
	pspDebugScreenInit();
	pspDebugScreenSetBackColor(0xffffffff);
	pspDebugScreenSetTextColor(0x00000000);
	pspDebugScreenClear();
	
	textbuffer = newTextBuffer();
	activebuffer = textbuffer;
	// Main Loop
	while (1){
	// 		Get data from pad
		sceCtrlReadBufferPositive(&pad,1);
		c = getChar(pad,charset);
	// 		Transform data from pad to actions
		//if (c < 0 && c != -20 && c != -39 && c != -58) while (pad.Buttons) sceCtrlReadBufferPositive(&pad,1); // Wait for all buttons to be released, so we act only one time
		switch (c){
			case -58:
			case -39:
			case -20:
			case 0: {
				if (oldc > 0){
					bufferAddChar(activebuffer,oldc);
					cursorx++;
				}
				break;
			}
			case -1: break;
			case -10: {
				while (pad.Buttons) sceCtrlReadBufferPositive(&pad,1);
				if (commandmode){
					execCommand(commandbuffer->lines->text, textbuffer);
					commandmode = 0;
					destroyBuffer(commandbuffer);
					activebuffer = textbuffer;
				}
				else{
					commandmode = 1;
					commandbuffer = newTextBuffer();
					activebuffer = commandbuffer;
				};
				break;
			}
			case -12: {
				while (pad.Buttons) sceCtrlReadBufferPositive(&pad,1);
				bufferAddChar(activebuffer,' ');
				cursorx++;
				break;
			}
			case -13: {
				while (pad.Buttons) sceCtrlReadBufferPositive(&pad,1);
				if (charset) charset = 0; // By now there are only two charsets implemented, so this is enough
				else charset = 1;
				break;
			}
			case -14: {
				while (pad.Buttons) sceCtrlReadBufferPositive(&pad,1);
				bufferDelChar(activebuffer);
				cursorx--;
				break;
			}
			case -15: {
				while (pad.Buttons) sceCtrlReadBufferPositive(&pad,1);
				bufferAddChar(activebuffer,'\n');
				//bufferNewLine(activebuffer);
				cursorx = 0;
				cursory++;
				break;
			}
			case -31: {
				while (pad.Buttons & PSP_CTRL_RIGHT) sceCtrlReadBufferPositive(&pad,1);
				moveCursorRight(activebuffer);
				break;
			}
			case -32: {
				while (pad.Buttons & PSP_CTRL_UP) sceCtrlReadBufferPositive(&pad,1);
				moveCursorUp(activebuffer);
				break;
			}
			case -33: {
				while (pad.Buttons & PSP_CTRL_LEFT) sceCtrlReadBufferPositive(&pad,1);
				moveCursorLeft(activebuffer);
				break;
			}
			case -34: {
				while (pad.Buttons & PSP_CTRL_DOWN) sceCtrlReadBufferPositive(&pad,1);
				moveCursorDown(activebuffer);
				break;
			}
		}
	// 		Clear and redraw screen if needed
		if (c != oldc){
			pspDebugScreenClear();
			printWriteHelp(pad,cursorx,cursory,charset);
			printBuffer(textbuffer);
			if (commandmode) printCommandBuffer(commandbuffer);
			oldc = c;
		}
	}
	return 0;
}
